@@ -18,10 +18,12 @@
 package ethstats
 
 import (
+	"context"
 	"encoding/json"
 	"errors"
 	"fmt"
 	"math/big"
+	"net"
 	"regexp"
 	"runtime"
 	"strconv"
@@ -29,31 +31,48 @@ import (
 	"time"
 
 	"github.com/ethereum/go-ethereum/common"
+	"github.com/ethereum/go-ethereum/common/mclock"
+	"github.com/ethereum/go-ethereum/consensus"
 	"github.com/ethereum/go-ethereum/core"
 	"github.com/ethereum/go-ethereum/core/types"
 	"github.com/ethereum/go-ethereum/eth"
 	"github.com/ethereum/go-ethereum/event"
 	"github.com/ethereum/go-ethereum/les"
-	"github.com/ethereum/go-ethereum/logger"
-	"github.com/ethereum/go-ethereum/logger/glog"
-	"github.com/ethereum/go-ethereum/node"
+	"github.com/ethereum/go-ethereum/log"
 	"github.com/ethereum/go-ethereum/p2p"
 	"github.com/ethereum/go-ethereum/rpc"
 	"golang.org/x/net/websocket"
 )
 
-// historyUpdateRange is the number of blocks a node should report upon login or
-// history request.
-const historyUpdateRange = 50
+const (
+	// historyUpdateRange is the number of blocks a node should report upon login or
+	// history request.
+	historyUpdateRange = 50
+
+	// txChanSize is the size of channel listening to NewTxsEvent.
+	// The number is referenced from the size of tx pool.
+	txChanSize = 4096
+	// chainHeadChanSize is the size of channel listening to ChainHeadEvent.
+	chainHeadChanSize = 10
+)
+
+type txPool interface {
+	// SubscribeNewTxsEvent should return an event subscription of
+	// NewTxsEvent and send events to the given channel.
+	SubscribeNewTxsEvent(chan<- core.NewTxsEvent) event.Subscription
+}
+
+type blockChain interface {
+	SubscribeChainHeadEvent(ch chan<- core.ChainHeadEvent) event.Subscription
+}
 
 // Service implements an Ethereum netstats reporting daemon that pushes local
 // chain statistics up to a monitoring server.
 type Service struct {
-	stack *node.Node // Temporary workaround, remove when API finalized
-
 	server *p2p.Server        // Peer-to-peer server to retrieve networking infos
 	eth    *eth.Ethereum      // Full Ethereum service if monitoring a full node
 	les    *les.LightEthereum // Light Ethereum service if monitoring a light node
+	engine consensus.Engine   // Consensus engine to retrieve variadic block fields
 
 	node string // Name of the node to display on the monitoring page
 	pass string // Password to authorize access to the monitoring page
@@ -72,9 +91,16 @@ func New(url string, ethServ *eth.Ethereum, lesServ *les.LightEthereum) (*Servic
 		return nil, fmt.Errorf("invalid netstats url: \"%s\", should be nodename:secret@host:port", url)
 	}
 	// Assemble and return the stats service
+	var engine consensus.Engine
+	if ethServ != nil {
+		engine = ethServ.Engine()
+	} else {
+		engine = lesServ.Engine()
+	}
 	return &Service{
 		eth:    ethServ,
 		les:    lesServ,
+		engine: engine,
 		node:   parts[1],
 		pass:   parts[3],
 		host:   parts[4],
@@ -96,64 +122,119 @@ func (s *Service) Start(server *p2p.Server) error {
 	s.server = server
 	go s.loop()
 
-	glog.V(logger.Info).Infoln("Stats daemon started")
+	log.Info("Stats daemon started")
 	return nil
 }
 
 // Stop implements node.Service, terminating the monitoring and reporting daemon.
 func (s *Service) Stop() error {
-	glog.V(logger.Info).Infoln("Stats daemon stopped")
+	log.Info("Stats daemon stopped")
 	return nil
 }
 
 // loop keeps trying to connect to the netstats server, reporting chain events
 // until termination.
 func (s *Service) loop() {
-	// Subscribe tso chain events to execute updates on
-	var emux *event.TypeMux
+	// Subscribe to chain events to execute updates on
+	var blockchain blockChain
+	var txpool txPool
 	if s.eth != nil {
-		emux = s.eth.EventMux()
+		blockchain = s.eth.BlockChain()
+		txpool = s.eth.TxPool()
 	} else {
-		emux = s.les.EventMux()
+		blockchain = s.les.BlockChain()
+		txpool = s.les.TxPool()
 	}
-	headSub := emux.Subscribe(core.ChainHeadEvent{})
+
+	chainHeadCh := make(chan core.ChainHeadEvent, chainHeadChanSize)
+	headSub := blockchain.SubscribeChainHeadEvent(chainHeadCh)
 	defer headSub.Unsubscribe()
 
-	txSub := emux.Subscribe(core.TxPreEvent{})
+	txEventCh := make(chan core.NewTxsEvent, txChanSize)
+	txSub := txpool.SubscribeNewTxsEvent(txEventCh)
 	defer txSub.Unsubscribe()
 
+	// Start a goroutine that exhausts the subsciptions to avoid events piling up
+	var (
+		quitCh = make(chan struct{})
+		headCh = make(chan *types.Block, 1)
+		txCh   = make(chan struct{}, 1)
+	)
+	go func() {
+		var lastTx mclock.AbsTime
+
+	HandleLoop:
+		for {
+			select {
+			// Notify of chain head events, but drop if too frequent
+			case head := <-chainHeadCh:
+				select {
+				case headCh <- head.Block:
+				default:
+				}
+
+			// Notify of new transaction events, but drop if too frequent
+			case <-txEventCh:
+				if time.Duration(mclock.Now()-lastTx) < time.Second {
+					continue
+				}
+				lastTx = mclock.Now()
+
+				select {
+				case txCh <- struct{}{}:
+				default:
+				}
+
+			// node stopped
+			case <-txSub.Err():
+				break HandleLoop
+			case <-headSub.Err():
+				break HandleLoop
+			}
+		}
+		close(quitCh)
+	}()
 	// Loop reporting until termination
 	for {
-		// Establish a websocket connection to the server and authenticate the node
-		url := fmt.Sprintf("%s/api", s.host)
-		if !strings.Contains(url, "://") {
-			url = "wss://" + url
+		// Resolve the URL, defaulting to TLS, but falling back to none too
+		path := fmt.Sprintf("%s/api", s.host)
+		urls := []string{path}
+
+		if !strings.Contains(path, "://") { // url.Parse and url.IsAbs is unsuitable (https://github.com/golang/go/issues/19779)
+			urls = []string{"wss://" + path, "ws://" + path}
+		}
+		// Establish a websocket connection to the server on any supported URL
+		var (
+			conf *websocket.Config
+			conn *websocket.Conn
+			err  error
+		)
+		for _, url := range urls {
+			if conf, err = websocket.NewConfig(url, "http://localhost/"); err != nil {
+				continue
+			}
+			conf.Dialer = &net.Dialer{Timeout: 5 * time.Second}
+			if conn, err = websocket.DialConfig(conf); err == nil {
+				break
+			}
 		}
-		conn, err := websocket.Dial(url, "", "http://localhost/")
 		if err != nil {
-			glog.V(logger.Warn).Infof("Stats server unreachable: %v", err)
+			log.Warn("Stats server unreachable", "err", err)
 			time.Sleep(10 * time.Second)
 			continue
 		}
-		in := json.NewDecoder(conn)
-		out := json.NewEncoder(conn)
-
-		if err = s.login(in, out); err != nil {
-			glog.V(logger.Warn).Infof("Stats login failed: %v", err)
+		// Authenticate the client with the server
+		if err = s.login(conn); err != nil {
+			log.Warn("Stats login failed", "err", err)
 			conn.Close()
 			time.Sleep(10 * time.Second)
 			continue
 		}
-		go s.readLoop(conn, in)
+		go s.readLoop(conn)
 
 		// Send the initial stats so our node looks decent from the get go
-		if err = s.report(out); err != nil {
-			glog.V(logger.Warn).Infof("Initial stats report failed: %v", err)
-			conn.Close()
-			continue
-		}
-		if err = s.reportHistory(out, nil); err != nil {
-			glog.V(logger.Warn).Infof("History report failed: %v", err)
+		if err = s.report(conn); err != nil {
+			log.Warn("Initial stats report failed", "err", err)
 			conn.Close()
 			continue
 		}
@@ -162,40 +243,28 @@ func (s *Service) loop() {
 
 		for err == nil {
 			select {
+			case <-quitCh:
+				conn.Close()
+				return
+
 			case <-fullReport.C:
-				if err = s.report(out); err != nil {
-					glog.V(logger.Warn).Infof("Full stats report failed: %v", err)
+				if err = s.report(conn); err != nil {
+					log.Warn("Full stats report failed", "err", err)
 				}
 			case list := <-s.histCh:
-				if err = s.reportHistory(out, list); err != nil {
-					glog.V(logger.Warn).Infof("Block history report failed: %v", err)
+				if err = s.reportHistory(conn, list); err != nil {
+					log.Warn("Requested history report failed", "err", err)
 				}
-			case head, ok := <-headSub.Chan():
-				if !ok { // node stopped
-					conn.Close()
-					return
+			case head := <-headCh:
+				if err = s.reportBlock(conn, head); err != nil {
+					log.Warn("Block stats report failed", "err", err)
 				}
-				if err = s.reportBlock(out, head.Data.(core.ChainHeadEvent).Block); err != nil {
-					glog.V(logger.Warn).Infof("Block stats report failed: %v", err)
+				if err = s.reportPending(conn); err != nil {
+					log.Warn("Post-block transaction stats report failed", "err", err)
 				}
-				if err = s.reportPending(out); err != nil {
-					glog.V(logger.Warn).Infof("Post-block transaction stats report failed: %v", err)
-				}
-			case _, ok := <-txSub.Chan():
-				if !ok { // node stopped
-					conn.Close()
-					return
-				}
-				// Exhaust events to avoid reporting too frequently
-				for exhausted := false; !exhausted; {
-					select {
-					case <-headSub.Chan():
-					default:
-						exhausted = true
-					}
-				}
-				if err = s.reportPending(out); err != nil {
-					glog.V(logger.Warn).Infof("Transaction stats report failed: %v", err)
+			case <-txCh:
+				if err = s.reportPending(conn); err != nil {
+					log.Warn("Transaction stats report failed", "err", err)
 				}
 			}
 		}
@@ -208,24 +277,25 @@ func (s *Service) loop() {
 // from the network socket. If any of them match an active request, it forwards
 // it, if they themselves are requests it initiates a reply, and lastly it drops
 // unknown packets.
-func (s *Service) readLoop(conn *websocket.Conn, in *json.Decoder) {
+func (s *Service) readLoop(conn *websocket.Conn) {
 	// If the read loop exists, close the connection
 	defer conn.Close()
 
 	for {
 		// Retrieve the next generic network packet and bail out on error
 		var msg map[string][]interface{}
-		if err := in.Decode(&msg); err != nil {
-			glog.V(logger.Warn).Infof("Failed to decode stats server message: %v", err)
+		if err := websocket.JSON.Receive(conn, &msg); err != nil {
+			log.Warn("Failed to decode stats server message", "err", err)
 			return
 		}
+		log.Trace("Received message from stats server", "msg", msg)
 		if len(msg["emit"]) == 0 {
-			glog.V(logger.Warn).Infof("Stats server sent non-broadcast: %v", msg)
+			log.Warn("Stats server sent non-broadcast", "msg", msg)
 			return
 		}
 		command, ok := msg["emit"][0].(string)
 		if !ok {
-			glog.V(logger.Warn).Infof("Invalid stats server message type: %v", msg["emit"][0])
+			log.Warn("Invalid stats server message type", "type", msg["emit"][0])
 			return
 		}
 		// If the message is a ping reply, deliver (someone must be listening!)
@@ -236,7 +306,7 @@ func (s *Service) readLoop(conn *websocket.Conn, in *json.Decoder) {
 				continue
 			default:
 				// Ping routine dead, abort
-				glog.V(logger.Warn).Infof("Stats server pinger seems to have died")
+				log.Warn("Stats server pinger seems to have died")
 				return
 			}
 		}
@@ -245,12 +315,13 @@ func (s *Service) readLoop(conn *websocket.Conn, in *json.Decoder) {
 			// Make sure the request is valid and doesn't crash us
 			request, ok := msg["emit"][1].(map[string]interface{})
 			if !ok {
-				glog.V(logger.Warn).Infof("Invalid history request: %v", msg["emit"][1])
-				return
+				log.Warn("Invalid stats history request", "msg", msg["emit"][1])
+				s.histCh <- nil
+				continue // Ethstats sometime sends invalid history requests, ignore those
 			}
 			list, ok := request["list"].([]interface{})
 			if !ok {
-				glog.V(logger.Warn).Infof("Invalid history block list: %v", request["list"])
+				log.Warn("Invalid stats history block list", "list", request["list"])
 				return
 			}
 			// Convert the block number list to an integer list
@@ -258,7 +329,7 @@ func (s *Service) readLoop(conn *websocket.Conn, in *json.Decoder) {
 			for i, num := range list {
 				n, ok := num.(float64)
 				if !ok {
-					glog.V(logger.Warn).Infof("Invalid history block number: %v", num)
+					log.Warn("Invalid stats history block number", "number", num)
 					return
 				}
 				numbers[i] = uint64(n)
@@ -270,7 +341,7 @@ func (s *Service) readLoop(conn *websocket.Conn, in *json.Decoder) {
 			}
 		}
 		// Report anything else and continue
-		glog.V(logger.Info).Infof("Unknown stats message: %v", msg)
+		log.Info("Unknown stats message", "msg", msg)
 	}
 }
 
@@ -291,26 +362,26 @@ type nodeInfo struct {
 
 // authMsg is the authentication infos needed to login to a monitoring server.
 type authMsg struct {
-	Id     string   `json:"id"`
+	ID     string   `json:"id"`
 	Info   nodeInfo `json:"info"`
 	Secret string   `json:"secret"`
 }
 
 // login tries to authorize the client at the remote server.
-func (s *Service) login(in *json.Decoder, out *json.Encoder) error {
+func (s *Service) login(conn *websocket.Conn) error {
 	// Construct and send the login authentication
 	infos := s.server.NodeInfo()
 
 	var network, protocol string
 	if info := infos.Protocols["eth"]; info != nil {
-		network = strconv.Itoa(info.(*eth.EthNodeInfo).Network)
+		network = fmt.Sprintf("%d", info.(*eth.NodeInfo).Network)
 		protocol = fmt.Sprintf("eth/%d", eth.ProtocolVersions[0])
 	} else {
-		network = strconv.Itoa(infos.Protocols["les"].(*eth.EthNodeInfo).Network)
-		protocol = fmt.Sprintf("les/%d", les.ProtocolVersions[0])
+		network = fmt.Sprintf("%d", infos.Protocols["les"].(*les.NodeInfo).Network)
+		protocol = fmt.Sprintf("les/%d", les.ClientProtocolVersions[0])
 	}
 	auth := &authMsg{
-		Id: s.node,
+		ID: s.node,
 		Info: nodeInfo{
 			Name:     s.node,
 			Node:     infos.Name,
@@ -326,14 +397,14 @@ func (s *Service) login(in *json.Decoder, out *json.Encoder) error {
 		Secret: s.pass,
 	}
 	login := map[string][]interface{}{
-		"emit": []interface{}{"hello", auth},
+		"emit": {"hello", auth},
 	}
-	if err := out.Encode(login); err != nil {
+	if err := websocket.JSON.Send(conn, login); err != nil {
 		return err
 	}
 	// Retrieve the remote ack or connection termination
 	var ack map[string][]string
-	if err := in.Decode(&ack); err != nil || len(ack["emit"]) != 1 || ack["emit"][0] != "ready" {
+	if err := websocket.JSON.Receive(conn, &ack); err != nil || len(ack["emit"]) != 1 || ack["emit"][0] != "ready" {
 		return errors.New("unauthorized")
 	}
 	return nil
@@ -342,17 +413,17 @@ func (s *Service) login(in *json.Decoder, out *json.Encoder) error {
 // report collects all possible data to report and send it to the stats server.
 // This should only be used on reconnects or rarely to avoid overloading the
 // server. Use the individual methods for reporting subscribed events.
-func (s *Service) report(out *json.Encoder) error {
-	if err := s.reportLatency(out); err != nil {
+func (s *Service) report(conn *websocket.Conn) error {
+	if err := s.reportLatency(conn); err != nil {
 		return err
 	}
-	if err := s.reportBlock(out, nil); err != nil {
+	if err := s.reportBlock(conn, nil); err != nil {
 		return err
 	}
-	if err := s.reportPending(out); err != nil {
+	if err := s.reportPending(conn); err != nil {
 		return err
 	}
-	if err := s.reportStats(out); err != nil {
+	if err := s.reportStats(conn); err != nil {
 		return err
 	}
 	return nil
@@ -360,63 +431,61 @@ func (s *Service) report(out *json.Encoder) error {
 
 // reportLatency sends a ping request to the server, measures the RTT time and
 // finally sends a latency update.
-func (s *Service) reportLatency(out *json.Encoder) error {
+func (s *Service) reportLatency(conn *websocket.Conn) error {
 	// Send the current time to the ethstats server
 	start := time.Now()
 
 	ping := map[string][]interface{}{
-		"emit": []interface{}{"node-ping", map[string]string{
+		"emit": {"node-ping", map[string]string{
 			"id":         s.node,
 			"clientTime": start.String(),
 		}},
 	}
-	if err := out.Encode(ping); err != nil {
+	if err := websocket.JSON.Send(conn, ping); err != nil {
 		return err
 	}
 	// Wait for the pong request to arrive back
 	select {
 	case <-s.pongCh:
 		// Pong delivered, report the latency
-	case <-time.After(3 * time.Second):
+	case <-time.After(5 * time.Second):
 		// Ping timeout, abort
 		return errors.New("ping timed out")
 	}
+	latency := strconv.Itoa(int((time.Since(start) / time.Duration(2)).Nanoseconds() / 1000000))
+
 	// Send back the measured latency
-	latency := map[string][]interface{}{
-		"emit": []interface{}{"latency", map[string]string{
+	log.Trace("Sending measured latency to ethstats", "latency", latency)
+
+	stats := map[string][]interface{}{
+		"emit": {"latency", map[string]string{
 			"id":      s.node,
-			"latency": strconv.Itoa(int((time.Since(start) / time.Duration(2)).Nanoseconds() / 1000000)),
+			"latency": latency,
 		}},
 	}
-	if err := out.Encode(latency); err != nil {
-		return err
-	}
-	return nil
+	return websocket.JSON.Send(conn, stats)
 }
 
 // blockStats is the information to report about individual blocks.
 type blockStats struct {
-	Number    *big.Int       `json:"number"`
-	Hash      common.Hash    `json:"hash"`
-	Timestamp *big.Int       `json:"timestamp"`
-	Miner     common.Address `json:"miner"`
-	GasUsed   *big.Int       `json:"gasUsed"`
-	GasLimit  *big.Int       `json:"gasLimit"`
-	Diff      string         `json:"difficulty"`
-	TotalDiff string         `json:"totalDifficulty"`
-	Txs       txStats        `json:"transactions"`
-	Uncles    uncleStats     `json:"uncles"`
+	Number     *big.Int       `json:"number"`
+	Hash       common.Hash    `json:"hash"`
+	ParentHash common.Hash    `json:"parentHash"`
+	Timestamp  *big.Int       `json:"timestamp"`
+	Miner      common.Address `json:"miner"`
+	GasUsed    uint64         `json:"gasUsed"`
+	GasLimit   uint64         `json:"gasLimit"`
+	Diff       string         `json:"difficulty"`
+	TotalDiff  string         `json:"totalDifficulty"`
+	Txs        []txStats      `json:"transactions"`
+	TxHash     common.Hash    `json:"transactionsRoot"`
+	Root       common.Hash    `json:"stateRoot"`
+	Uncles     uncleStats     `json:"uncles"`
 }
 
-// txStats is a custom wrapper around a transaction array to force serializing
-// empty arrays instead of returning null for them.
-type txStats []*types.Transaction
-
-func (s txStats) MarshalJSON() ([]byte, error) {
-	if txs := ([]*types.Transaction)(s); len(txs) > 0 {
-		return json.Marshal(txs)
-	}
-	return []byte("[]"), nil
+// txStats is the information to report about individual transactions.
+type txStats struct {
+	Hash common.Hash `json:"hash"`
 }
 
 // uncleStats is a custom wrapper around an uncle array to force serializing
@@ -431,19 +500,21 @@ func (s uncleStats) MarshalJSON() ([]byte, error) {
 }
 
 // reportBlock retrieves the current chain head and repors it to the stats server.
-func (s *Service) reportBlock(out *json.Encoder, block *types.Block) error {
-	// Assemble the block stats report and send it to the server
+func (s *Service) reportBlock(conn *websocket.Conn, block *types.Block) error {
+	// Gather the block details from the header or block chain
+	details := s.assembleBlockStats(block)
+
+	// Assemble the block report and send it to the server
+	log.Trace("Sending new block to ethstats", "number", details.Number, "hash", details.Hash)
+
 	stats := map[string]interface{}{
 		"id":    s.node,
-		"block": s.assembleBlockStats(block),
+		"block": details,
 	}
 	report := map[string][]interface{}{
-		"emit": []interface{}{"block", stats},
+		"emit": {"block", stats},
 	}
-	if err := out.Encode(report); err != nil {
-		return err
-	}
-	return nil
+	return websocket.JSON.Send(conn, report)
 }
 
 // assembleBlockStats retrieves any required metadata to report a single block
@@ -453,7 +524,7 @@ func (s *Service) assembleBlockStats(block *types.Block) *blockStats {
 	var (
 		header *types.Header
 		td     *big.Int
-		txs    []*types.Transaction
+		txs    []txStats
 		uncles []*types.Header
 	)
 	if s.eth != nil {
@@ -464,7 +535,10 @@ func (s *Service) assembleBlockStats(block *types.Block) *blockStats {
 		header = block.Header()
 		td = s.eth.BlockChain().GetTd(header.Hash(), header.Number.Uint64())
 
-		txs = block.Transactions()
+		txs = make([]txStats, len(block.Transactions()))
+		for i, tx := range block.Transactions() {
+			txs[i].Hash = tx.Hash()
+		}
 		uncles = block.Uncles()
 	} else {
 		// Light nodes would need on-demand lookups for transactions/uncles, skip
@@ -474,69 +548,87 @@ func (s *Service) assembleBlockStats(block *types.Block) *blockStats {
 			header = s.les.BlockChain().CurrentHeader()
 		}
 		td = s.les.BlockChain().GetTd(header.Hash(), header.Number.Uint64())
+		txs = []txStats{}
 	}
 	// Assemble and return the block stats
+	author, _ := s.engine.Author(header)
+
 	return &blockStats{
-		Number:    header.Number,
-		Hash:      header.Hash(),
-		Timestamp: header.Time,
-		Miner:     header.Coinbase,
-		GasUsed:   new(big.Int).Set(header.GasUsed),
-		GasLimit:  new(big.Int).Set(header.GasLimit),
-		Diff:      header.Difficulty.String(),
-		TotalDiff: td.String(),
-		Txs:       txs,
-		Uncles:    uncles,
+		Number:     header.Number,
+		Hash:       header.Hash(),
+		ParentHash: header.ParentHash,
+		Timestamp:  header.Time,
+		Miner:      author,
+		GasUsed:    header.GasUsed,
+		GasLimit:   header.GasLimit,
+		Diff:       header.Difficulty.String(),
+		TotalDiff:  td.String(),
+		Txs:        txs,
+		TxHash:     header.TxHash,
+		Root:       header.Root,
+		Uncles:     uncles,
 	}
 }
 
 // reportHistory retrieves the most recent batch of blocks and reports it to the
 // stats server.
-func (s *Service) reportHistory(out *json.Encoder, list []uint64) error {
+func (s *Service) reportHistory(conn *websocket.Conn, list []uint64) error {
 	// Figure out the indexes that need reporting
 	indexes := make([]uint64, 0, historyUpdateRange)
 	if len(list) > 0 {
 		// Specific indexes requested, send them back in particular
-		for _, idx := range list {
-			indexes = append(indexes, idx)
-		}
+		indexes = append(indexes, list...)
 	} else {
 		// No indexes requested, send back the top ones
-		var head *types.Header
+		var head int64
 		if s.eth != nil {
-			head = s.eth.BlockChain().CurrentHeader()
+			head = s.eth.BlockChain().CurrentHeader().Number.Int64()
 		} else {
-			head = s.les.BlockChain().CurrentHeader()
+			head = s.les.BlockChain().CurrentHeader().Number.Int64()
 		}
-		start := head.Number.Int64() - historyUpdateRange
+		start := head - historyUpdateRange + 1
 		if start < 0 {
 			start = 0
 		}
-		for i := uint64(start); i <= head.Number.Uint64(); i++ {
+		for i := uint64(start); i <= uint64(head); i++ {
 			indexes = append(indexes, i)
 		}
 	}
 	// Gather the batch of blocks to report
 	history := make([]*blockStats, len(indexes))
 	for i, number := range indexes {
+		// Retrieve the next block if it's known to us
+		var block *types.Block
 		if s.eth != nil {
-			history[i] = s.assembleBlockStats(s.eth.BlockChain().GetBlockByNumber(number))
+			block = s.eth.BlockChain().GetBlockByNumber(number)
 		} else {
-			history[i] = s.assembleBlockStats(types.NewBlockWithHeader(s.les.BlockChain().GetHeaderByNumber(number)))
+			if header := s.les.BlockChain().GetHeaderByNumber(number); header != nil {
+				block = types.NewBlockWithHeader(header)
+			}
 		}
+		// If we do have the block, add to the history and continue
+		if block != nil {
+			history[len(history)-1-i] = s.assembleBlockStats(block)
+			continue
+		}
+		// Ran out of blocks, cut the report short and send
+		history = history[len(history)-i:]
+		break
 	}
 	// Assemble the history report and send it to the server
+	if len(history) > 0 {
+		log.Trace("Sending historical blocks to ethstats", "first", history[0].Number, "last", history[len(history)-1].Number)
+	} else {
+		log.Trace("No history to send to stats server")
+	}
 	stats := map[string]interface{}{
 		"id":      s.node,
 		"history": history,
 	}
 	report := map[string][]interface{}{
-		"emit": []interface{}{"history", stats},
+		"emit": {"history", stats},
 	}
-	if err := out.Encode(report); err != nil {
-		return err
-	}
-	return nil
+	return websocket.JSON.Send(conn, report)
 }
 
 // pendStats is the information to report about pending transactions.
@@ -546,7 +638,7 @@ type pendStats struct {
 
 // reportPending retrieves the current number of pending transactions and reports
 // it to the stats server.
-func (s *Service) reportPending(out *json.Encoder) error {
+func (s *Service) reportPending(conn *websocket.Conn) error {
 	// Retrieve the pending count from the local blockchain
 	var pending int
 	if s.eth != nil {
@@ -555,6 +647,8 @@ func (s *Service) reportPending(out *json.Encoder) error {
 		pending = s.les.TxPool().Stats()
 	}
 	// Assemble the transaction stats and send it to the server
+	log.Trace("Sending pending transactions to ethstats", "count", pending)
+
 	stats := map[string]interface{}{
 		"id": s.node,
 		"stats": &pendStats{
@@ -562,15 +656,12 @@ func (s *Service) reportPending(out *json.Encoder) error {
 		},
 	}
 	report := map[string][]interface{}{
-		"emit": []interface{}{"pending", stats},
+		"emit": {"pending", stats},
 	}
-	if err := out.Encode(report); err != nil {
-		return err
-	}
-	return nil
+	return websocket.JSON.Send(conn, report)
 }
 
-// blockStats is the information to report about the local node.
+// nodeStats is the information to report about the local node.
 type nodeStats struct {
 	Active   bool `json:"active"`
 	Syncing  bool `json:"syncing"`
@@ -583,7 +674,7 @@ type nodeStats struct {
 
 // reportPending retrieves various stats about the node at the networking and
 // mining layer and reports it to the stats server.
-func (s *Service) reportStats(out *json.Encoder) error {
+func (s *Service) reportStats(conn *websocket.Conn) error {
 	// Gather the syncing and mining infos from the local miner instance
 	var (
 		mining   bool
@@ -598,11 +689,15 @@ func (s *Service) reportStats(out *json.Encoder) error {
 		sync := s.eth.Downloader().Progress()
 		syncing = s.eth.BlockChain().CurrentHeader().Number.Uint64() >= sync.HighestBlock
 
-		gasprice = int(s.eth.Miner().GasPrice().Uint64())
+		price, _ := s.eth.APIBackend.SuggestPrice(context.Background())
+		gasprice = int(price.Uint64())
 	} else {
 		sync := s.les.Downloader().Progress()
 		syncing = s.les.BlockChain().CurrentHeader().Number.Uint64() >= sync.HighestBlock
 	}
+	// Assemble the node stats and send it to the server
+	log.Trace("Sending node details to ethstats")
+
 	stats := map[string]interface{}{
 		"id": s.node,
 		"stats": &nodeStats{
@@ -616,10 +711,7 @@ func (s *Service) reportStats(out *json.Encoder) error {
 		},
 	}
 	report := map[string][]interface{}{
-		"emit": []interface{}{"stats", stats},
+		"emit": {"stats", stats},
 	}
-	if err := out.Encode(report); err != nil {
-		return err
-	}
-	return nil
+	return websocket.JSON.Send(conn, report)
 }
